#version 150
precision highp float; precision highp int;
//Ethan Alexander Shulman 2022

in vec3 in_Position;
out vec4 uv;

layout(std140) uniform ShaderviewBlock {
	vec4 mouse;
	float screenX,screenY;
	int frame,framerate;
};

vec4 baseRand;
float nsampling;

#include "art/PathTracerScene.glsl"

vec3 normal(vec3 p) {
	nsampling = 1.;
	vec3 n = normalize(vec3(geometry(p+vec3(LIGHTING_EPSILON,0,0)).x,geometry(p+vec3(0,LIGHTING_EPSILON,0)).x,geometry(p+vec3(0,0,LIGHTING_EPSILON)).x)-geometry(p).x);
	nsampling = 0.;
	if (n != n) return vec3(0,1,0);
	return n;
}

vec3 rp, rd, c = vec3(1), l = vec3(0);
bool refractState = false, refracted = false;

void initialBounce(vec2 uv, vec3 nrm) {
	vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)),
		bitang = cross(tang,nrm);
	mat3 mat = material(rp,nrm,geometry(rp).y);
	vec2 rand;
	
	//low discrepancy sequence keeps errors in high frequency to aid denoising extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
	#if !defined(LOW_DISCREPANCY) || DENOISE_RANGE == 0
	rand = hash(uvec4(abs(rp.xyz)*10000.,frame)).xy*I2F;
	#else
	#define DENOISE_TOTAL (DENOISE_RANGE*2+1)
	#ifdef ANIMATION_TIME
	int seqId = frame%ANIMATION_SAMPLES;
	#else
	int seqId = frame;
	#endif
	seqId = seqId*DENOISE_TOTAL*DENOISE_TOTAL+int(dot(mod(floor(uv),float(DENOISE_TOTAL)),vec2(1,DENOISE_TOTAL)));
	const float r2 = 1.32471795724474602596;
	rand = fract(0.5+seqId*(1./vec2(r2*r2,r2)));
	#endif
		
	float ct = pow(rand.x,mat[1].x), st = sqrt(1.-ct*ct),
		pt = rand.y*PI*2.;
	vec3 bump = normalize(nrm*ct + tang*(st*cos(pt)) + bitang*(st*sin(pt)));
	if (fract(rand.x+rand.x*rand.y*1000.) > mat[1].z) {
		rd = refract(rd,bump,refractState?mat[2].x:1./mat[2].x);
		rp -= nrm*(LIGHTING_EPSILON*2.);
		refractState = !refractState;
		refracted = true;
	} else {
		rd = reflect(rd,bump);
		rp += nrm*(LIGHTING_EPSILON*2.);
	}
}

void bounce(vec3 nrm, float matId) {
	vec3 tang = normalize(cross(abs(nrm.y)>(1.-1e-5)?vec3(0,0,-1):vec3(0,-1,0),nrm)),
		bitang = cross(tang,nrm);
	mat3 mat = material(rp,nrm,matId);
	c *= mat[0];
	#ifdef COLOR_MAX
	float clen = length(c);
	if (clen > COLOR_MAX) c *= COLOR_MAX/clen;
	#endif
	l += mat[1].y*c;
	
	vec4 rand = hash(uvec4(abs(rp.xyz)*10000.,frame))*I2F;
	baseRand = rand;
	float ct = pow(rand.x,mat[1].x), st = sqrt(1.-ct*ct),
		pt = rand.y*PI*2.;
	vec3 bump = normalize(nrm*ct + tang*(st*cos(pt)) + bitang*(st*sin(pt)));
	if (rand.z > mat[1].z) {
		rd = refract(rd,bump,refractState?mat[2].x:1./mat[2].x);
		rp -= nrm*(LIGHTING_EPSILON*2.);
		refractState = !refractState;
	} else {
		rd = reflect(rd,bump);
		rp += nrm*(LIGHTING_EPSILON*2.);
	}
}

uniform sampler2D Buffer0;

void vert() {
	#ifdef ANIMATION_TIME
	time = mod(float(frame)/float(framerate*ANIMATION_SAMPLES),ANIMATION_TIME);
	#endif
	uv = vec4(in_Position.xy*vec2(screenX/screenY,1),(in_Position.xy*.5+.5)*vec2(screenX,screenY));
	gl_Position = vec4(in_Position,1);
}
void frag() {
	//setup camera ray
	mat3 cam = camera();
	vec4 frand = hash(uvec4(frame)*uvec4(1,10,100,1000))*I2F;
	rp = vec3(0,0,cam[2].z);
	rd = normalize(vec3(uv.xy+(frand.xy*2.-1.)/screenY,cam[2].x));
	if (cam[1].x != 0.) {
		mat2 m = r2d(cam[1].x);
		rp.xy *= m; rd.xy *= m;
	}
	if (cam[1].y != 0.) {
		mat2 m = r2d(cam[1].y);
		rp.yz *= m; rd.yz *= m;
	}
	if (cam[1].z != 0.) {
		mat2 m = r2d(cam[1].z);
		rp.xz *= m; rd.xz *= m;
	}
	rp += cam[0].xyz;

	//sample view buffer
	vec4 viewHit = texelFetch(Buffer0,ivec2(uv.zw),0);
	if (viewHit.w >= 0.) {
		//step forward to hit
		rp += rd*viewHit.w;
		
		//initial ray bounce
		baseRand = frand;
	 	initialBounce(uv.zw, viewHit.xyz);
	
		//path trace
		#ifdef LIGHTING_STEP_DITHER
		float dither = (1.-LIGHTING_STEP_DITHER)+LIGHTING_STEP_DITHER*frand.z;
		#endif
		for (int i = 0; i < LIGHTING_STEPS; i++) {
			vec2 geo = geometry(rp);
			if (geo.x < LIGHTING_EPSILON) {
				vec3 nrm = normal(rp-rd*LIGHTING_EPSILON);
				bounce(nrm,geo.y);
				#ifdef LIGHTING_STEP_DITHER
				float dither = (1.-LIGHTING_STEP_DITHER)+LIGHTING_STEP_DITHER*baseRand.w;
				#endif
			} else {
				#ifdef LIGHTING_STEP_DITHER
				rp += rd*(geo.x*dither);
				#else
				rp += rd*geo.x;
				#endif
			}
		}
		l += c*backgroundLighting(rp,rd);
		gl_FragColor = vec4(l,refracted?-1.:1.);
		
	} else {
		gl_FragColor = vec4(background(rp,rd),0);
	}
}
